from kura.base_classes import BaseClusteringMethod
from sklearn.cluster import KMeans
import math
from typing import TypeVar
import numpy as np

T = TypeVar("T")


class KmeansClusteringMethod(BaseClusteringMethod):
    def __init__(self, clusters_per_group: int = 10):
        self.clusters_per_group = clusters_per_group

    def cluster(self, items: list[T]) -> dict[int, list[T]]:
        """
        We perform a clustering here using an embedding defined on each individual item.

        We assume that the item is passed in as a dictionary with

        - its relevant embedding stored in the "embedding" key.
        - the item itself stored in the "item" key.

        {
            "embedding": list[float],
            "item": any,
        }
        """

        embeddings = [item["embedding"] for item in items]  # pyright: ignore
        data: list[T] = [item["item"] for item in items]  # pyright: ignore
        n_clusters = math.ceil(len(data) / self.clusters_per_group)

        X = np.array(embeddings)
        kmeans = KMeans(n_clusters=n_clusters)
        cluster_labels = kmeans.fit_predict(X)

        return {
            i: [data[j] for j in range(len(data)) if cluster_labels[j] == i]
            for i in range(n_clusters)
        }



from kura.base_classes import BaseClusterModel, BaseClusteringMethod, BaseEmbeddingModel
from kura.embedding import OpenAIEmbeddingModel
from kura.k_means import KmeansClusteringMethod
from kura.types import ConversationSummary, Cluster, GeneratedCluster
from tqdm.asyncio import tqdm_asyncio
import numpy as np
from asyncio import Semaphore
import instructor
import google.generativeai as genai
import os


class ClusterModel(BaseClusterModel):
    def __init__(
        self,
        clustering_method: BaseClusteringMethod = KmeansClusteringMethod(),
        embedding_model: BaseEmbeddingModel = OpenAIEmbeddingModel(),
        max_concurrent_requests: int = 50,
        client=instructor.from_gemini(
            genai.GenerativeModel("gemini-1.5-flash-latest"), use_async=True
        ),
    ):
        self.clustering_method = clustering_method
        self.embedding_model = embedding_model
        self.max_concurrent_requests = max_concurrent_requests
        self.client = client

    def get_contrastive_examples(
        self,
        cluster_id: int,
        cluster_id_to_summaries: dict[int, list[ConversationSummary]],
        desired_count: int = 10,
    ):
        other_clusters = [c for c in cluster_id_to_summaries.keys() if c != cluster_id]
        all_examples = []
        for cluster in other_clusters:
            all_examples.extend(cluster_id_to_summaries[cluster])

        # If we don't have enough examples, return all of them
        if len(all_examples) <= desired_count:
            return all_examples

        # Otherwise sample without replacement
        return list(np.random.choice(all_examples, size=desired_count, replace=False))

    async def generate_cluster(
        self,
        summaries: list[ConversationSummary],
        contrastive_examples: list[ConversationSummary],
        sem: Semaphore,
    ) -> Cluster:
        async with sem:
            resp = await self.client.chat.completions.create(
                messages=[
                    {
                        "role": "system",
                        "content": """
                    You are tasked with summarizing a group of related statements into a short, precise, and accurate description and name. Your goal is to create a concise summary that captures the essence of these statements and distinguishes them from other similar groups of statements.
                    
                    Summarize all the statements into a clear, precise, two-sentence description in the past tense. Your summary should be specific to this group and distinguish it from the contrastive answers of the other groups.

                    After creating the summary, generate a short name for the group of statements. This name should be at most ten words long (perhaps less) and be specific but also reflective of most of the statements (rather than reflecting only one or two). The name should distinguish this group from the contrastive examples. For instance, "Write fantasy sexual roleplay with octopi and monsters", "Generate blog spam for gambling websites", or "Assist with high school math homework" would be better and more actionable than general terms like "Write erotic content" or "Help with homework". Be as descriptive as possible and assume neither good nor bad faith. Do not hesitate to identify and describe socially harmful or sensitive topics specifically; specificity is necessary for monitoring
                    
                    The cluster name should be a sentence in the imperative that captures the user’s request. For example, ‘Brainstorm ideas for a birthday party’ or ‘Help me find a new job.”
                    """,
                    },
                    {
                        "role": "user",
                        "content": """
                    Here are the relevant statements

                    Below are the related statements:
                    <positive_examples>
                    {% for item in positive_examples %}
                        <positive_example>{{ item.summary }}</positive_example>
                    {% endfor %}
                    </positive_examples>  
                    
                    For context, here are statements from nearby groups that are NOT part of the group you’re summarizing
                    
                    <contrastive_examples>
                    {% for item in contrastive_examples %}
                    <contrastive_example>{{ item.summary }}</contrastive_example>
                    {% endfor %}
                    </contrastive_examples>
                    
                    Remember to analyze both the statements and the contrastive statements carefully to ensure your summary and name accurately represent the specific group while distinguishing it from others.
                    """,
                    },
                ],
                response_model=GeneratedCluster,
                context={
                    "positive_examples": summaries,
                    "contrastive_examples": contrastive_examples,
                },
            )

            return Cluster(
                name=resp.name,
                description=resp.summary,
                chat_ids=[item.chat_id for item in summaries],
                parent_id=None,
            )

    async def cluster_summaries(
        self, summaries: list[ConversationSummary]
    ) -> list[Cluster]:
        sem = Semaphore(self.max_concurrent_requests)
        embeddings: list[list[float]] = await tqdm_asyncio.gather(
            *[
                self.embedding_model.embed(text=item.summary, sem=sem)
                for item in summaries
            ],
            desc="Embedding Summaries",
        )
        cluster_id_to_summaries = self.clustering_method.cluster(
            [
                {
                    "item": item,
                    "embedding": embedding,
                }
                for item, embedding in zip(summaries, embeddings)
            ]
        )
        clusters: list[Cluster] = await tqdm_asyncio.gather(
            *[
                self.generate_cluster(
                    summaries,
                    self.get_contrastive_examples(
                        cluster_id, cluster_id_to_summaries, 10
                    ),
                    sem,
                )
                for cluster_id, summaries in cluster_id_to_summaries.items()
            ],
            desc="Generating Base Clusters",
        )

        return clusters



from abc import ABC, abstractmethod
from kura.types.cluster import Cluster


class BaseMetaClusterModel(ABC):
    @abstractmethod
    async def reduce_clusters(self, clusters: list[Cluster]) -> list[Cluster]:
        pass



from kura.base_classes import BaseSummaryModel
from kura.types import Conversation, ConversationSummary
from kura.types.summarisation import GeneratedSummary
from asyncio import Semaphore
from tqdm.asyncio import tqdm_asyncio
import google.generativeai as genai
import instructor
import os


class SummaryModel(BaseSummaryModel):
    def __init__(
        self,
        max_concurrent_requests: int = 50,
    ):
        self.sem = Semaphore(max_concurrent_requests)
        self.client = instructor.from_gemini(
            genai.GenerativeModel(
                model_name="gemini-1.5-flash-latest",
            ),
            use_async=True,
        )

    async def summarise(
        self, conversations: list[Conversation]
    ) -> list[ConversationSummary]:
        summaries = await tqdm_asyncio.gather(
            *[
                self.summarise_conversation(conversation)
                for conversation in conversations
            ],
            desc=f"Summarising {len(conversations)} conversations",
        )
        return summaries

    async def apply_hooks(
        self, conversation: ConversationSummary
    ) -> ConversationSummary:
        # TODO: Implement hooks here for extra metadata extraction
        return await super().apply_hooks(conversation)

    async def summarise_conversation(
        self, conversation: Conversation
    ) -> ConversationSummary:
        resp = await self.client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": """
                    Generate a summary of the task that the user is asking the language model to do based off the following conversation.


                    The summary should be concise and short. It should be at most 1-2 sentences and at most 30 words. Here are some examples of summaries:
                    - The user’s overall request for the assistant is to help implementing a React component to display a paginated list of users from a database.
                    - The user’s overall request for the assistant is to debug a memory leak in their Python data processing pipeline.
                    - The user’s overall request for the assistant is to design and architect a REST API for a social media application.
                    

                    Here is the conversation
                    <messages>
                    {% for message in messages %}
                        <message>{{message.role}}: {{message.content}} </message>
                    {% endfor %}
                    </messages>

                    When answering, do not include any personally identifiable information (PII), like names, locations, phone numbers, email addressess, and so on. When answering, do not include any proper nouns. Make sure that you're clear, concise and that you get to the point in at most two sentences.
                    
                    For example:

                    Remember that
                    - Summaries should be concise and short. They should each be at most 1-2 sentences and at most 30 words.
                    - Summaries should start with "The user's overall request for the assistant is to"
                    - Make sure to omit any personally identifiable information (PII), like names, locations, phone numbers, email addressess, company names and so on.
                    - Make sure to indicate specific details such as programming languages, frameworks, libraries and so on which are relevant to the task.
                    """,
                }
            ],
            context={"messages": conversation.messages},
            response_model=GeneratedSummary,
        )
        return ConversationSummary(
            chat_id=conversation.chat_id,
            summary=resp.summary,
            metadata={
                "conversation_turns": len(conversation.messages),
            },
        )



from pydantic import BaseModel
from datetime import datetime
from typing import Literal
import json


class Message(BaseModel):
    created_at: datetime
    role: Literal["user", "assistant"]
    content: str


class Conversation(BaseModel):
    chat_id: str
    created_at: datetime
    messages: list[Message]

    @classmethod
    def from_claude_conversation_dump(cls, file_path: str) -> list["Conversation"]:
        with open(file_path, "r") as f:
            return [
                Conversation(
                    chat_id=conversation["uuid"],
                    created_at=conversation["created_at"],
                    messages=[
                        Message(
                            created_at=datetime.fromisoformat(
                                message["created_at"].replace("Z", "+00:00")
                            ),
                            role="user"
                            if message["sender"] == "human"
                            else "assistant",
                            content="\n".join(
                                [
                                    item["text"]
                                    for item in message["content"]
                                    if item["type"] == "text"
                                ]
                            ),
                        )
                        for message in sorted(
                            conversation["chat_messages"],
                            key=lambda x: (
                                datetime.fromisoformat(
                                    x["created_at"].replace("Z", "+00:00")
                                ),
                                0 if x["sender"] == "human" else 1,
                            ),
                        )
                    ],
                )
                for conversation in json.load(f)
            ]



from kura.base_classes import BaseEmbeddingModel
from asyncio import Semaphore, wait_for
from tenacity import retry, wait_fixed, stop_after_attempt
from openai import AsyncOpenAI


class OpenAIEmbeddingModel(BaseEmbeddingModel):
    def __init__(self):
        self.client = AsyncOpenAI()

    @retry(wait=wait_fixed(3), stop=stop_after_attempt(3))
    async def embed(self, text: str, sem: Semaphore) -> list[float]:
        async with sem:
            resp = await wait_for(
                self.client.embeddings.create(
                    input=text, model="text-embedding-3-small"
                ),
                timeout=5,
            )
            return resp.data[0].embedding



from pydantic import BaseModel, Field, computed_field
import uuid
from typing import Union


class Cluster(BaseModel):
    id: str = Field(
        default_factory=lambda: uuid.uuid4().hex,
    )
    name: str
    description: str
    chat_ids: list[str]
    parent_id: Union[str, None]

    @computed_field
    def count(self) -> int:
        return len(self.chat_ids)


class GeneratedCluster(BaseModel):
    name: str
    summary: str



from abc import ABC, abstractmethod

from kura.types import Conversation, ConversationSummary


class BaseSummaryModel(ABC):
    @abstractmethod
    async def summarise(
        self, conversations: list[Conversation]
    ) -> list[ConversationSummary]:
        """Summarise the conversations into a list of ConversationSummary"""
        pass

    @abstractmethod
    async def summarise_conversation(
        self, conversation: Conversation
    ) -> ConversationSummary:
        """Summarise a single conversation into a single string"""
        pass

    @abstractmethod
    async def apply_hooks(
        self, conversation: ConversationSummary
    ) -> ConversationSummary:
        """Apply hooks to the conversation summary"""
        # Assert that the implementation of the class has a hooks attribute so we can call it in summarise_conversation
        assert hasattr(self, "hooks")
        pass



from abc import ABC, abstractmethod
from asyncio import Semaphore


class BaseEmbeddingModel(ABC):
    @abstractmethod
    async def embed(self, text: str, sem: Semaphore) -> list[float]:
        """Embed a single text into a list of floats"""
        pass



from kura.dimensionality import HDBUMAP
from kura.types import Conversation, Cluster
from kura.embedding import OpenAIEmbeddingModel
from kura.summarisation import SummaryModel
from kura.meta_cluster import MetaClusterModel
from kura.cluster import ClusterModel
from kura.base_classes import (
    BaseEmbeddingModel,
    BaseSummaryModel,
    BaseClusterModel,
    BaseMetaClusterModel,
    BaseDimensionalityReduction,
)
from pathlib import Path
from typing import Union
import os
from typing import TypeVar
from pydantic import BaseModel

from kura.types.dimensionality import ProjectedCluster
from kura.types.summarisation import ConversationSummary

T = TypeVar("T", bound=BaseModel)


class Kura:
    def __init__(
        self,
        embedding_model: BaseEmbeddingModel = OpenAIEmbeddingModel(),
        summarisation_model: BaseSummaryModel = SummaryModel(),
        cluster_model: BaseClusterModel = ClusterModel(),
        meta_cluster_model: BaseMetaClusterModel = MetaClusterModel(),
        dimensionality_reduction: BaseDimensionalityReduction = HDBUMAP(),
        max_clusters: int = 10,
        checkpoint_dir: str = "./checkpoints",
        summary_checkpoint_name: str = "summaries.jsonl",
        cluster_checkpoint_name: str = "clusters.jsonl",
        meta_cluster_checkpoint_name: str = "meta_clusters.jsonl",
        dimensionality_checkpoint_name: str = "dimensionality.jsonl",
        disable_checkpoints: bool = False,
    ):
        # Define Models that we're using
        self.embedding_model = embedding_model
        self.embedding_model = embedding_model
        self.summarisation_model = summarisation_model
        self.max_clusters = max_clusters
        self.cluster_model = cluster_model
        self.meta_cluster_model = meta_cluster_model
        self.dimensionality_reduction = dimensionality_reduction

        # Define Checkpoints
        self.checkpoint_dir = os.path.join(checkpoint_dir)
        self.cluster_checkpoint_name = os.path.join(
            self.checkpoint_dir, cluster_checkpoint_name
        )
        self.meta_cluster_checkpoint_name = os.path.join(
            self.checkpoint_dir, meta_cluster_checkpoint_name
        )
        self.dimensionality_checkpoint_name = os.path.join(
            self.checkpoint_dir, dimensionality_checkpoint_name
        )
        self.summary_checkpoint_name = os.path.join(
            self.checkpoint_dir, summary_checkpoint_name
        )
        self.disable_checkpoints = disable_checkpoints

        if not os.path.exists(self.checkpoint_dir) and not self.disable_checkpoints:
            os.makedirs(self.checkpoint_dir)

        if not self.disable_checkpoints:
            print(f"Checkpoint directory: {Path(self.checkpoint_dir)}")

    def load_checkpoint(
        self, checkpoint_path: str, response_model: type[T]
    ) -> Union[list[T], None]:
        if not self.disable_checkpoints:
            if os.path.exists(checkpoint_path):
                print(
                    f"Loading checkpoint from {checkpoint_path} for {response_model.__name__}"
                )
                with open(checkpoint_path, "r") as f:
                    return [response_model.model_validate_json(line) for line in f]
        return None

    def save_checkpoint(self, checkpoint_path: str, data: list[T]) -> None:
        if not self.disable_checkpoints:
            with open(checkpoint_path, "w") as f:
                for item in data:
                    f.write(item.model_dump_json() + "\n")

    async def reduce_clusters(self, clusters: list[Cluster]) -> list[Cluster]:
        checkpoint_items = self.load_checkpoint(
            self.meta_cluster_checkpoint_name, Cluster
        )
        if checkpoint_items:
            return checkpoint_items

        root_clusters = clusters

        print(f"Starting with {len(root_clusters)} clusters")

        while len(root_clusters) > self.max_clusters:
            # We get the updated list of clusters
            new_current_level = await self.meta_cluster_model.reduce_clusters(
                root_clusters
            )

            # These are the new root clusters that we've generated
            root_clusters = [c for c in new_current_level if c.parent_id is None]

            # We then remove outdated versions of clusters
            old_cluster_ids = {rc.id for rc in new_current_level if rc.parent_id}
            clusters = [c for c in clusters if c.id not in old_cluster_ids]

            # We then add the new clusters to the list
            clusters.extend(new_current_level)

            print(f"Reduced to {len(root_clusters)} clusters")

        self.save_checkpoint(self.meta_cluster_checkpoint_name, clusters)
        return clusters

    async def summarise_conversations(
        self, conversations: list[Conversation]
    ) -> list[ConversationSummary]:
        checkpoint_items = self.load_checkpoint(
            self.summary_checkpoint_name, ConversationSummary
        )
        if checkpoint_items:
            return checkpoint_items

        summaries = await self.summarisation_model.summarise(conversations)
        self.save_checkpoint(self.summary_checkpoint_name, summaries)
        return summaries

    async def generate_base_clusters(self, summaries: list[ConversationSummary]):
        base_cluster_checkpoint_items = self.load_checkpoint(
            self.cluster_checkpoint_name, Cluster
        )
        if base_cluster_checkpoint_items:
            return base_cluster_checkpoint_items

        clusters: list[Cluster] = await self.cluster_model.cluster_summaries(summaries)
        self.save_checkpoint(self.cluster_checkpoint_name, clusters)
        return clusters

    async def reduce_dimensionality(
        self, clusters: list[Cluster]
    ) -> list[ProjectedCluster]:
        checkpoint_items = self.load_checkpoint(
            self.dimensionality_checkpoint_name, ProjectedCluster
        )
        if checkpoint_items:
            return checkpoint_items

        dimensionality_reduced_clusters = (
            await self.dimensionality_reduction.reduce_dimensionality(clusters)
        )

        self.save_checkpoint(
            self.dimensionality_checkpoint_name, dimensionality_reduced_clusters
        )
        return dimensionality_reduced_clusters

    async def cluster_conversations(self, conversations: list[Conversation]):
        summaries = await self.summarise_conversations(conversations)
        clusters: list[Cluster] = await self.generate_base_clusters(summaries)
        processed_clusters: list[Cluster] = await self.reduce_clusters(clusters)
        dimensionality_reduced_clusters = await self.reduce_dimensionality(
            processed_clusters
        )

        return dimensionality_reduced_clusters



from fastapi import FastAPI, staticfiles
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from pathlib import Path
from kura import Kura
from kura.types import ProjectedCluster, Conversation
from typing import Optional
from kura.cli.visualisation import (
    generate_cumulative_chart_data,
    generate_messages_per_chat_data,
    generate_messages_per_week_data,
    generate_new_chats_per_week_data,
)
import json


api = FastAPI()

# Configure CORS
api.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)

# Serve static files from web/dist
web_dir = Path(__file__).parent.parent / "static" / "dist"
if not web_dir.exists():
    raise FileNotFoundError(f"Static files directory not found: {web_dir}")


class ConversationData(BaseModel):
    data: list[Conversation]
    max_clusters: Optional[int]
    disable_checkpoints: bool


@api.post("/api/analyse")
async def analyse_conversations(conversation_data: ConversationData):
    print(conversation_data.disable_checkpoints)
    # Load clusters from checkpoint file if it exists
    clusters_file = Path("./checkpoints") / "dimensionality.jsonl"
    if not clusters_file.exists() or conversation_data.disable_checkpoints:
        kura = Kura(
            checkpoint_dir=str(clusters_file.parent),
            max_clusters=conversation_data.max_clusters
            if conversation_data.max_clusters
            else 10,
            disable_checkpoints=conversation_data.disable_checkpoints,
        )
        clusters = await kura.cluster_conversations(conversation_data.data)
    else:
        with open(clusters_file) as f:
            clusters_data = []
            for line in f:
                clusters_data.append(line)
            clusters = [
                ProjectedCluster(**json.loads(cluster)) for cluster in clusters_data
            ]

    return {
        "cumulative_words": generate_cumulative_chart_data(conversation_data.data),
        "messages_per_chat": generate_messages_per_chat_data(conversation_data.data),
        "messages_per_week": generate_messages_per_week_data(conversation_data.data),
        "new_chats_per_week": generate_new_chats_per_week_data(conversation_data.data),
        "clusters": clusters,
    }


# Serve static files from web/dist at the root
web_dir = Path(__file__).parent.parent / "static" / "dist"
if not web_dir.exists():
    raise FileNotFoundError(f"Static files directory not found: {web_dir}")

# Mount static files at root
api.mount("/", staticfiles.StaticFiles(directory=str(web_dir), html=True))



import typer
import uvicorn
from kura.cli.server import api
from rich import print
import os

app = typer.Typer()


@app.command()
def start_app(
    dir: str = typer.Option(
        "./checkpoints",
        help="Directory to use for checkpoints, relative to the current directory",
    ),
):
    """Start the FastAPI server"""
    os.environ["KURA_CHECKPOINT_DIR"] = dir
    uvicorn.run(api, host="0.0.0.0", port=8000)
    print(
        "\n[bold green]🚀 Access website at[/bold green] [bold blue][http://localhost:8000](http://localhost:8000)[/bold blue]\n"
    )


if __name__ == "__main__":
    app()



from kura.base_classes import BaseDimensionalityReduction, BaseEmbeddingModel
from kura.types import Cluster, ProjectedCluster
from kura.embedding import OpenAIEmbeddingModel
from umap import UMAP
import numpy as np
import asyncio
import os
from numpy.typing import NDArray
from numpy import float64


class HDBUMAP(BaseDimensionalityReduction):
    def __init__(
        self,
        embedding_model: BaseEmbeddingModel = OpenAIEmbeddingModel(),
    ):
        self.embedding_model = embedding_model

    async def reduce_dimensionality(
        self, clusters: list[Cluster]
    ) -> list[ProjectedCluster]:
        # Embed all clusters
        sem = asyncio.Semaphore(50)
        cluster_embeddings = await asyncio.gather(
            *[
                self.embedding_model.embed(
                    f"Name: {c.name}\nDescription: {c.description}", sem
                )
                for c in clusters
            ]
        )

        # Convert embeddings to numpy array
        embeddings = np.array(cluster_embeddings)

        # Project to 2D using UMAP
        umap_reducer = UMAP(
            n_components=2,
            n_neighbors=min(15, len(embeddings) - 1),
            min_dist=0,
            metric="cosine",
        )
        reduced_embeddings = umap_reducer.fit_transform(embeddings)

        # Create projected clusters with 2D coordinates
        res = []
        for i, cluster in enumerate(clusters):
            projected = ProjectedCluster(
                id=cluster.id,
                name=cluster.name,
                description=cluster.description,
                chat_ids=cluster.chat_ids,
                parent_id=cluster.parent_id,
                x_coord=float(reduced_embeddings[i][0]),  # pyright: ignore
                y_coord=float(reduced_embeddings[i][1]),  # pyright: ignore
                level=0 if cluster.parent_id is None else 1,
            )
            res.append(projected)

        return res



from abc import ABC, abstractmethod
from kura.types import ConversationSummary, Cluster


class BaseClusterModel(ABC):
    @abstractmethod
    async def cluster_summaries(
        self, summaries: list[ConversationSummary]
    ) -> list[Cluster]:
        pass

    # TODO : Add abstract method for hooks here once we start supporting it



from .embedding import BaseEmbeddingModel
from .summarisation import BaseSummaryModel
from .clustering_method import BaseClusteringMethod
from .cluster import BaseClusterModel
from .meta_cluster import BaseMetaClusterModel
from .dimensionality import BaseDimensionalityReduction

__all__ = [
    "BaseEmbeddingModel",
    "BaseSummaryModel",
    "BaseClusteringMethod",
    "BaseClusterModel",
    "BaseMetaClusterModel",
    "BaseDimensionalityReduction",
]



from abc import ABC, abstractmethod
from typing import TypeVar, Union

T = TypeVar("T")


class BaseClusteringMethod(ABC):
    @abstractmethod
    def cluster(
        self, items: list[dict[str, Union[T, list[float]]]]
    ) -> dict[int, list[T]]:
        pass



from abc import ABC, abstractmethod

from kura.types import Cluster, ProjectedCluster


class BaseDimensionalityReduction(ABC):
    @abstractmethod
    async def reduce_dimensionality(
        self, clusters: list[Cluster]
    ) -> list[ProjectedCluster]:
        """
        This reduces the dimensionality of the individual clusters that we've created so we can visualise them in a lower dimension
        """
        pass



from .conversation import Conversation, Message
from .summarisation import ConversationSummary, GeneratedSummary
from .cluster import Cluster, GeneratedCluster
from .dimensionality import ProjectedCluster

__all__ = [
    "Cluster",
    "Conversation",
    "Message",
    "ConversationSummary",
    "GeneratedSummary",
    "GeneratedCluster",
    "ProjectedCluster",
]
